/****************************************************************************
**
*W  system.c                    GAP source                       Frank Celler
*W                                                         & Martin Schönert
*W                                                         & Dave Bayer (MAC)
*W                                                  & Harald Boegeholz (OS/2)
*W                                                         & Paul Doyle (VMS)
*W                                                  & Burkhard Höfling (MAC)
*W                                                    & Steve Linton (MS/DOS)
**
**
*Y  Copyright (C)  1996,  Lehrstuhl D für Mathematik,  RWTH Aachen,  Germany
*Y  (C) 1998 School Math and Comp. Sci., University of St Andrews, Scotland
*Y  Copyright (C) 2002 The GAP Group
**
**  The  files   "system.c" and  "sysfiles.c"  contains all  operating system
**  dependent  functions.  This file contains  all system dependent functions
**  except file and stream operations, which are implemented in "sysfiles.c".
**  The following labels determine which operating system is actually used.
**
**  Under UNIX autoconf  is used to check  various features of  the operating
**  system and the compiler.  Should you have problem compiling GAP check the
**  file "bin/CPU-VENDOR-OS/config.h" after you have done a
**
**     ./configure ; make config
**
**  in the root directory.  And then do a
**
**     make compile
**
**  to compile and link GAP.
*/

#include        "system.h"              /* system dependent part           */



#include        "sysfiles.h"            /* file input/output               */
#include        "gasman.h"            
#include        <fcntl.h>


#include        <stdio.h>               /* standard input/output functions */
#include        <stdlib.h>              /* ANSI standard functions         */
#include        <string.h>              /* string functions                */

#include        <assert.h>
#include        <dirent.h>

#if HAVE_UNISTD_H                       /* definition of 'R_OK'            */
# include        <unistd.h>
#endif


#if HAVE_LIBREADLINE
#include        <readline/readline.h>   /* readline for interactive input  */
#endif

#if HAVE_SYS_TIMES_H                    /* time functions                  */
# include       <sys/types.h>
# include       <sys/times.h>
#endif

#if HAVE_MADVISE
#include        <sys/mman.h>
#endif

/****************************************************************************
**
*V  SyKernelVersion  . . . . . . . . . . . . . . . . name of the architecture
** do not edit the following line. Occurences of `4.7.5' and `today'
** will be replaced by string matching by distribution wrapping scripts.
*/
const libGAP_Char * libGAP_SyKernelVersion = "4.7.5";

/****************************************************************************
*V  SyWindowsPath  . . . . . . . . . . . . . . . . . default path for Windows
** do not edit the following line. Occurences of `gap4r7'
** will be replaced by string matching by distribution wrapping scripts.
*/
const libGAP_Char * libGAP_SyWindowsPath = "/cygdrive/c/gap4r7";

/****************************************************************************
**

*F * * * * * * * * * * * command line settable options  * * * * * * * * * * *
*/

/****************************************************************************
**

*V  SyStackAlign  . . . . . . . . . . . . . . . . . .  alignment of the stack
**
**  'SyStackAlign' is  the  alignment  of items on the stack.   It  must be a
**  divisor of  'sizof(Bag)'.  The  addresses of all identifiers on the stack
**  must be  divisable by 'SyStackAlign'.  So if it  is 1, identifiers may be
**  anywhere on the stack, and if it is  'sizeof(Bag)',  identifiers may only
**  be  at addresses  divisible by  'sizeof(Bag)'.  This value is initialized
**  from a macro passed from the makefile, because it is machine dependent.
**
**  This value is passed to 'InitBags'.
*/
libGAP_UInt libGAP_SyStackAlign = libGAP_SYS_STACK_ALIGN;


/****************************************************************************
**
*V  SyArchitecture  . . . . . . . . . . . . . . . .  name of the architecture
*/
const libGAP_Char * libGAP_SyArchitecture = SYS_ARCH;


/****************************************************************************
**
*V  SyCTRD  . . . . . . . . . . . . . . . . . . .  true if '<ctr>-D' is <eof>
*/
libGAP_UInt libGAP_SyCTRD;


/****************************************************************************
**
*V  SyCacheSize . . . . . . . . . . . . . . . . . . . . . . size of the cache
**
**  'SyCacheSize' is the size of the data cache.
**
**  This is per  default 0, which means that  there is no usuable data cache.
**  It is usually changed with the '-c' option in the script that starts GAP.
**
**  This value is passed to 'InitBags'.
**
**  Put in this package because the command line processing takes place here.
*/
libGAP_UInt libGAP_SyCacheSize;


/****************************************************************************
**
*V  SyCheckCRCCompiledModule  . . .  check crc while loading compiled modules
*/
libGAP_Int libGAP_SyCheckCRCCompiledModule;


/****************************************************************************
**
*V  SyCompileInput  . . . . . . . . . . . . . . . . . .  from this input file
*/
libGAP_Char libGAP_SyCompileInput [256];


/****************************************************************************
**
*V  SyCompileMagic1 . . . . . . . . . . . . . . . . . . and this magic string
*/
libGAP_Char * libGAP_SyCompileMagic1;


/****************************************************************************
**
*V  SyCompileName . . . . . . . . . . . . . . . . . . . . . .  with this name
*/
libGAP_Char libGAP_SyCompileName [256];


/****************************************************************************
**
*V  SyCompileOutput . . . . . . . . . . . . . . . . . . into this output file
*/
libGAP_Char libGAP_SyCompileOutput [256];

/****************************************************************************
**
*V  SyCompileOutput . . . . . . . . . . . . . . . . . . into this output file
*/
libGAP_Char libGAP_SyCompileOptions [256] = {'\0'};


/****************************************************************************
**
*V  SyCompilePlease . . . . . . . . . . . . . . .  tell GAP to compile a file
*/
libGAP_Int libGAP_SyCompilePlease;


/****************************************************************************
**
*V  SyDebugLoading  . . . . . . . . .  output messages about loading of files
*/
libGAP_Int libGAP_SyDebugLoading;


/****************************************************************************
**
*V  SyGapRootPaths  . . . . . . . . . . . . . . . . . . . array of root paths
**
**  'SyGapRootPaths' contains the  names   of the directories where   the GAP
**  files are located.
**
**  It is modified by the command line option -l.
**
**  It  is copied into the GAP  variable  called 'GAP_ROOT_PATHS' and used by
**  'SyFindGapRootFile'.
**
**  Each entry must end  with the pathname seperator, eg.  if 'init.g' is the
**  name of a library file 'strcat( SyGapRootPaths[i], "lib/init.g" );'  must
**  be a valid filename.
**
**  In addition we store the path to the users ~/.gap directory, if available,
**  in 'DotGapPath'.
**  
**  Put in this package because the command line processing takes place here.
**
#define MAX_GAP_DIRS 128
*/
libGAP_Char libGAP_SyGapRootPaths [libGAP_MAX_GAP_DIRS] [512];
#if HAVE_DOTGAPRC
libGAP_Char libGAP_DotGapPath[512];
#endif

/****************************************************************************
**
*V  IgnoreGapRC . . . . . . . . . . . . . . . . . . . -r option for kernel
**
*/
libGAP_Int libGAP_IgnoreGapRC;

/****************************************************************************
**
*V  SyHasUserHome . . . . . . . . . .  true if user has HOME in environment
*V  SyUserHome . . . . . . . . . . . . .  path of users home (it is exists)
*/
libGAP_Int libGAP_SyHasUserHome;
libGAP_Char libGAP_SyUserHome [256];

/****************************************************************************
**
*V  SyLineEdit  . . . . . . . . . . . . . . . . . . . .  support line editing
**
**  0: no line editing
**  1: line editing if terminal
**  2: always line editing (EMACS)
*/
libGAP_UInt libGAP_SyLineEdit;

/****************************************************************************
**
*V  SyUseReadline   . . . . . . . . . . . . . . . . . .  support line editing
**
**  Switch for not using readline although GAP is compiled with libreadline
*/
libGAP_UInt libGAP_SyUseReadline;

/****************************************************************************
**
*V  SyMsgsFlagBags  . . . . . . . . . . . . . . . . .  enable gasman messages
**
**  'SyMsgsFlagBags' determines whether garabage collections are reported  or
**  not.
**
**  Per default it is false, i.e. Gasman is silent about garbage collections.
**  It can be changed by using the  '-g'  option  on the  GAP  command  line.
**
**  This is used in the function 'SyMsgsBags' below.
**
**  Put in this package because the command line processing takes place here.
*/
libGAP_UInt libGAP_SyMsgsFlagBags;


/****************************************************************************
**
*V  SyNrCols  . . . . . . . . . . . . . . . . . .  length of the output lines
**
**  'SyNrCols' is the length of the lines on the standard output  device.
**
**  Per default this is 80 characters which is the usual width of  terminals.
**  It can be changed by the '-x' options for larger terminals  or  printers.
**
**  'Pr' uses this to decide where to insert a <newline> on the output lines.
**  'SyRead' uses it to decide when to start scrolling the echoed input line.
**
**  See also getwindowsize() below.
**
**  Put in this package because the command line processing takes place here.
*/
libGAP_UInt libGAP_SyNrCols;
libGAP_UInt libGAP_SyNrColsLocked;

/****************************************************************************
**
*V  SyNrRows  . . . . . . . . . . . . . . . . . number of lines on the screen
**
**  'SyNrRows' is the number of lines on the standard output device.
**
**  Per default this is 24, which is the  usual  size  of  terminal  screens.
**  It can be changed with the '-y' option for larger terminals or  printers.
**
**  'SyHelp' uses this to decide where to stop with '-- <space> for more --'.
**
**  See also getwindowsize() below.
*/
libGAP_UInt libGAP_SyNrRows;
libGAP_UInt libGAP_SyNrRowsLocked;

/****************************************************************************
**
*V  SyQuiet . . . . . . . . . . . . . . . . . . . . . . . . . suppress prompt
**
**  'SyQuiet' determines whether GAP should print the prompt and the  banner.
**
**  Per default its false, i.e. GAP prints the prompt and  the  nice  banner.
**  It can be changed by the '-q' option to have GAP operate in silent  mode.
**
**  It is used by the functions in 'gap.c' to suppress printing the  prompts.
**
**  Put in this package because the command line processing takes place here.
*/
libGAP_UInt libGAP_SyQuiet;


/****************************************************************************
**
*V  SyRestoring . . . . . . . . . . . . . . . . . . . . restoring a workspace
**
**  `SyRestoring' determines whether GAP is restoring a workspace or not.  If
**  it is zero no restoring should take place otherwise it holds the filename
**  of a workspace to restore.
**
*/
libGAP_Char * libGAP_SyRestoring;


/****************************************************************************
**
*V  SyInitializing                               set to 1 during library init
**
**  `SyInitializing' is set to 1 during the library intialization phase of
**  startup. It supresses some ebhaviours that may not be possible so early
**  such as homogeneity tests in the plist code.
*/

libGAP_UInt libGAP_SyInitializing;


/****************************************************************************
**
*V  SyStorMax . . . . . . . . . . . . . . . . . . . maximal size of workspace
**
**  'SyStorMax' is the maximal size of the workspace allocated by Gasman.
**  in kilobytes
**
**  This is per default 1G in 32-bit mode and 2G in 64-bit mode, which
**  is often a reasonable value. It is usually changed with the '-o'
**  option in the script that starts GAP.
**
**  This is used in the function 'SyAllocBags'below.
**
**  Put in this package because the command line processing takes place here.
*/
libGAP_Int libGAP_SyStorMax;
libGAP_Int libGAP_SyStorOverrun;

/****************************************************************************
**
*V  SyStorKill . . . . . . . . . . . . . . . . . . maximal size of workspace
**
**  'SyStorKill' is really the maximal size of the workspace allocated by 
**  Gasman. GAP exists before trying to allocate more than this amount
**  of memory.
**
**  This is per default disabled (i.e. = 0).
**  Can be changed with the '-K' option in the script that starts GAP.
**
**  This is used in the function 'SyAllocBags'below.
**
**  Put in this package because the command line processing takes place here.
*/
libGAP_Int libGAP_SyStorKill;


/****************************************************************************
**
*V  SyStorMin . . . . . . . . . . . . . .  default size for initial workspace
**
**  'SyStorMin' is the size of the initial workspace allocated by Gasman.
**
**  This is per default  24 Megabyte,  which  is  often  a  reasonable  value.
**  It is usually changed with the '-m' option in the script that starts GAP.
**
**  This value is used in the function 'SyAllocBags' below.
**
**  Put in this package because the command line processing takes place here.
*/
libGAP_Int libGAP_SyStorMin;


/****************************************************************************
**
*V  SySystemInitFile  . . . . . . . . . . .  name of the system "init.g" file
*/
libGAP_Char libGAP_SySystemInitFile [256];


/****************************************************************************
**
*V  SyUseModule . . . . . check for dynamic/static modules in 'READ_GAP_ROOT'
*/
int libGAP_SyUseModule;


/****************************************************************************
**
*V  SyWindow  . . . . . . . . . . . . . . . .  running under a window handler
**
**  'SyWindow' is 1 if GAP  is running under  a window handler front end such
**  as 'xgap', and 0 otherwise.
**
**  If running under  a window handler front  end, GAP adds various  commands
**  starting with '@' to the output to let 'xgap' know what is going on.
*/
libGAP_UInt libGAP_SyWindow;


/****************************************************************************
**

*F * * * * * * * * * * * * * time related functions * * * * * * * * * * * * *
*/

/****************************************************************************
**

*V  SyStartTime . . . . . . . . . . . . . . . . . . time when GAP was started
*/
libGAP_UInt libGAP_SyStartTime;


/****************************************************************************
**
*V  SyStopTime  . . . . . . . . . . . . . . . . . . time when reading started
*/
libGAP_UInt libGAP_SyStopTime;


/****************************************************************************
**
*F  SyTime()  . . . . . . . . . . . . . . . return time spent in milliseconds
**
**  'SyTime' returns the number of milliseconds spent by GAP so far.
**
**  Should be as accurate as possible,  because it  is  used  for  profiling.
*/


/****************************************************************************
**
*f  SyTime()  . . . . . . . . . . . . . . . . . . . . . . . using `getrusage'
**
**  Use use   'getrusage'  if possible,  because  it gives  us  a much better
**  resolution. 
*/
#if HAVE_GETRUSAGE && !SYS_IS_CYGWIN32

#if HAVE_SYS_TIME_H
# include       <sys/time.h>            /* definition of 'struct timeval'  */
#endif
#if HAVE_SYS_RESOURCE_H
# include       <sys/resource.h>        /* definition of 'struct rusage'   */
#endif

libGAP_UInt libGAP_SyTime ( void )
{
    struct rusage       buf;

    if ( getrusage( RUSAGE_SELF, &buf ) ) {
        fputs("gap: panic 'SyTime' cannot get time!\n",stderr);
        libGAP_SyExit( 1 );
    }
    return buf.ru_utime.tv_sec*1000 + buf.ru_utime.tv_usec/1000 - libGAP_SyStartTime;
}
libGAP_UInt libGAP_SyTimeSys ( void )
{
    struct rusage       buf;

    if ( getrusage( RUSAGE_SELF, &buf ) ) {
        fputs("gap: panic 'SyTimeSys' cannot get time!\n",stderr);
        libGAP_SyExit( 1 );
    }
    return buf.ru_stime.tv_sec*1000 + buf.ru_stime.tv_usec/1000;
}
libGAP_UInt libGAP_SyTimeChildren ( void )
{
    struct rusage       buf;

    if ( getrusage( RUSAGE_CHILDREN, &buf ) ) {
        fputs("gap: panic 'SyTimeChildren' cannot get time!\n",stderr);
        libGAP_SyExit( 1 );
    }
    return buf.ru_utime.tv_sec*1000 + buf.ru_utime.tv_usec/1000;
}
libGAP_UInt libGAP_SyTimeChildrenSys ( void )
{
    struct rusage       buf;

    if ( getrusage( RUSAGE_CHILDREN, &buf ) ) {
        fputs("gap: panic 'SyTimeChildrenSys' cannot get time!\n",stderr);
        libGAP_SyExit( 1 );
    }
    return buf.ru_stime.tv_sec*1000 + buf.ru_stime.tv_usec/1000;
}

#endif


/****************************************************************************
**
*f  SyTime()
**
**  The 'times' POSIX call measures clock ticks in 1/CLOCKS_PER_SEC. Typically,
**  CLOCKS_PER_SEC equals 60 or 100. On most systems, the 'times' API has been
**  deprecated in favor of 'getrusage', which provides a much better resolution.
**
**  Cygwin claims to support 'getrusage' but child times are not given properly.
*/
#if HAVE_TIMES && (SYS_IS_CYGWIN32 || !HAVE_GETRUSAGE)

#include <time.h> /* for CLOCKS_PER_SEC */

libGAP_UInt libGAP_SyTime ( void )
{
    struct tms          tbuf;

    if ( times( &tbuf ) == -1 ) {
        fputs("gap: panic 'SyTime' cannot get time!\n",stderr);
        libGAP_SyExit( 1 );
    }
    return tbuf.tms_utime * 1000L / CLOCKS_PER_SEC - libGAP_SyStartTime;
}

libGAP_UInt libGAP_SyTimeSys ( void )
{
    struct tms          tbuf;

    if ( times( &tbuf ) == -1 ) {
        fputs("gap: panic 'SyTimeSys' cannot get time!\n",stderr);
        libGAP_SyExit( 1 );
    }
    return tbuf.tms_stime * 1000L / CLOCKS_PER_SEC;
}

libGAP_UInt libGAP_SyTimeChildren ( void )
{
    struct tms          tbuf;

    if ( times( &tbuf ) == -1 ) {
        fputs("gap: panic 'SyTimeChildren' cannot get time!\n",stderr);
        libGAP_SyExit( 1 );
    }
    return tbuf.tms_cutime * 1000L / CLOCKS_PER_SEC;
}

libGAP_UInt libGAP_SyTimeChildrenSys ( void )
{
    struct tms          tbuf;

    if ( times( &tbuf ) == -1 ) {
        fputs("gap: panic 'SyTimeChildrenSys' cannot get time!\n",stderr);
        libGAP_SyExit( 1 );
    }
    return tbuf.tms_cstime * 1000L / CLOCKS_PER_SEC;
}

#endif


/****************************************************************************
**

*F * * * * * * * * * * * * * * * string functions * * * * * * * * * * * * * *
*/


/****************************************************************************
**
*F  SyStrlen( <str> ) . . . . . . . . . . . . . . . . . .  length of a string
**
**  'SyStrlen' returns the length of the string <str>, i.e.,  the  number  of
**  characters in <str> that precede the terminating null character.
*/
libGAP_UInt libGAP_SyStrlen (
    const libGAP_Char *         str )
{
    return strlen( str );
}


/****************************************************************************
**
*F  SyStrcmp( <str1>, <str2> )  . . . . . . . . . . . . . compare two strings
**
**  'SyStrcmp' returns an integer greater than, equal to, or less  than  zero
**  according to whether <str1> is greater  than,  equal  to,  or  less  than
**  <str2> lexicographically.
*/
libGAP_Int libGAP_SyStrcmp (
    const libGAP_Char *        str1,
    const libGAP_Char *        str2 )
{
    return strcmp( str1, str2 );
}


/****************************************************************************
**
*F  SyStrncmp( <str1>, <str2>, <len> )  . . . . . . . . . compare two strings
**
**  'SyStrncmp' returns an integer greater than, equal to,  or less than zero
**  according  to whether  <str1>  is greater than,  equal  to,  or less than
**  <str2> lexicographically.  'SyStrncmp' compares at most <len> characters.
*/
libGAP_Int libGAP_SyStrncmp (
    const libGAP_Char *        str1,
    const libGAP_Char *        str2,
    libGAP_UInt                len )
{
    return strncmp( str1, str2, len );
}

/****************************************************************************
**
*F  SyIntString( <string> ) . . . . . . . . extract a C integer from a string
**
*/




#if HAVE_ATOL
libGAP_Int libGAP_SyIntString( const libGAP_Char *string) {
  return atol (string);
}
#else
libGAP_Int libGAP_SyIntString( const libGAP_Char *string) {
  libGAP_Int x = 0;
  libGAP_Int sign = 1;
  while (libGAP_IsSpace(*string))
    string++;
  if (*string == '-')
    {
      sign = -1;
      string++;
    }
  else if (*string == '+')
    {
      string++;
    }
  while (libGAP_IsDigit(*string)) {
    x *= 10;
    x += (*string - '0');
    string++;
  }
  return sign*x;
}


#endif



/****************************************************************************
**
*F  SyStrncat( <dst>, <src>, <len> )  . . . . .  append one string to another
**
**  'SyStrncat'  appends characters from the  <src>  to <dst>  until either a
**  null character  is  encoutered  or  <len>  characters have   been copied.
**  <dst> becomes the concatenation of <dst> and <src>.  The resulting string
**  is always null terminated.  'SyStrncat' returns a pointer to <dst>.
*/
libGAP_Char * libGAP_SyStrncat (
    libGAP_Char *              dst,
    const libGAP_Char *        src,
    libGAP_UInt                len )
{
    return strncat( dst, src, len );
}


#ifndef HAVE_STRLCPY

size_t libGAP_strlcpy (
    char *dst,
    const char *src,
    size_t len)
{
    /* Keep a copy of the original src. */
    const char * const orig_src = src;

    /* If a non-empty len was specified, we can actually copy some data. */
    if (len > 0) {
        /* Copy up to len-1 bytes (reserve one for the terminating zero). */
        while (--len > 0) {
            /* Copy from src to dst; if we reach the string end, we are
               done and can simply return the total source string length */
            if ((*dst++ = *src++) == 0) {
                /* return length of source string without the zero byte */
                return src - orig_src - 1;
            }
        }
        
        /* If we got here, then we used up the whole buffer and len is zero.
           We must make sure to terminate the destination string. */
        *dst = 0;
    }

    /* in the end, we must return the length of the source string, no
       matter whether we completely copied or not; so advance src
       till its terminator is reached */
    while (*src++)
        ;

    /* return length of source string without the zero byte */
    return src - orig_src - 1;
}

#endif /* !HAVE_STRLCPY */


#ifndef HAVE_STRLCAT

size_t libGAP_strlcat (
    char *dst,
    const char *src,
    size_t len)
{
    /* Keep a copy of the original dst. */
    const char * const orig_dst = dst;

    /* Find the end of the dst string, so that we can append after it. */
    while (*dst != 0 && len > 0) {
        dst++;
        len--;
    }

    /* We can only append anything if there is free space left in the
       destination buffer. */
    if (len > 0) {
        /* One byte goes away for the terminating zero. */
        len--;

        /* Do the actual work and append from src to dst, until we either
           appended everything, or reached the dst buffer's end. */
        while (*src != 0 && len > 0) {
            *dst++ = *src++;
            len--;
        }

        /* Terminate, terminate, terminate! */
        *dst = 0;
    }

    /* Compute the final result. */
    return (dst - orig_dst) + strlen(src);
}

#endif /* !HAVE_STRLCAT */

size_t libGAP_strlncat (
    char *dst,
    const char *src,
    size_t len,
    size_t n)
{
    /* Keep a copy of the original dst. */
    const char * const orig_dst = dst;

    /* Find the end of the dst string, so that we can append after it. */
    while (*dst != 0 && len > 0) {
        dst++;
        len--;
    }

    /* We can only append anything if there is free space left in the
       destination buffer. */
    if (len > 0) {
        /* One byte goes away for the terminating zero. */
        len--;

        /* Do the actual work and append from src to dst, until we either
           appended everything, or reached the dst buffer's end. */
        while (*src != 0 && len > 0 && n > 0) {
            *dst++ = *src++;
            len--;
            n--;
        }

        /* Terminate, terminate, terminate! */
        *dst = 0;
    }

    /* Compute the final result. */
    len = strlen(src);
    if (n < len)
        len = n;
    return (dst - orig_dst) + len;
}

size_t libGAP_strxcpy (
    char *dst,
    const char *src,
    size_t len)
{
    size_t res = libGAP_strlcpy(dst, src, len);
    assert(res < len);
    return res;
}

size_t libGAP_strxcat (
    char *dst,
    const char *src,
    size_t len)
{
    size_t res = libGAP_strlcat(dst, src, len);
    assert(res < len);
    return res;
}

size_t libGAP_strxncat (
    char *dst,
    const char *src,
    size_t len,
    size_t n)
{
    size_t res = libGAP_strlncat(dst, src, len, n);
    assert(res < len);
    return res;
}

/****************************************************************************
**

*F * * * * * * * * * * * * * * gasman interface * * * * * * * * * * * * * * *
*/


/****************************************************************************
**
*F  SyMsgsBags( <full>, <phase>, <nr> ) . . . . . . . display Gasman messages
**
**  'SyMsgsBags' is the function that is used by Gasman to  display  messages
**  during garbage collections.
*/

libGAP_Int libGAP_SyGasmanNumbers[2][9];

void libGAP_SyMsgsBags (
    libGAP_UInt                full,
    libGAP_UInt                phase,
    libGAP_Int                 nr )
{
    libGAP_Char                cmd [3];        /* command string buffer           */
    libGAP_Char                str [32];       /* string buffer                   */
    libGAP_Char                ch;             /* leading character               */
    libGAP_UInt                i;              /* loop variable                   */
    libGAP_Int                 copynr;         /* copy of <nr>                    */
    libGAP_UInt                shifted;        /* non-zero if nr > 10^6 and so
                                           has to be shifted down          */
    static libGAP_UInt         tstart = 0;

    /* remember the numbers */
    if (phase > 0)
      {
        libGAP_SyGasmanNumbers[full][phase] = nr;
        
        /* in a full GC clear the partial numbers */
        if (full)
          libGAP_SyGasmanNumbers[0][phase] = 0;
      }
    else
      {
        libGAP_SyGasmanNumbers[full][0]++;
        tstart = libGAP_SyTime();
      }
    if (phase == 6) 
      {
        libGAP_UInt x = libGAP_SyTime() - tstart;
        libGAP_SyGasmanNumbers[full][7] = x;
        libGAP_SyGasmanNumbers[full][8] += x;
      }

    /* convert <nr> into a string with leading blanks                      */
    copynr = nr;
    ch = '0';  str[7] = '\0';
    shifted = (nr >= ((phase % 2) ? 10000000 : 1000000)) ? 1 : 0;
    if (shifted)
      {
        nr /= 1024;
      }
    if ((phase % 2) == 1 && shifted && nr > 1000000)
      {
        shifted++;
        nr /= 1024;
      }
      
    for ( i = ((phase % 2) == 1 && shifted) ? 6 : 7 ;
          i != 0; i-- ) {
        if      ( 0 < nr ) { str[i-1] = '0' + ( nr) % 10;  ch = ' '; }
        else if ( nr < 0 ) { str[i-1] = '0' + (-nr) % 10;  ch = '-'; }
        else               { str[i-1] = ch;                ch = ' '; }
        nr = nr / 10;
    }
    nr = copynr;

    if ((phase % 2) == 1 && shifted == 1)
      str[6] = 'K';
    if ((phase % 2) == 1 && shifted == 2)
      str[6] = 'M';

    

    /* ordinary full garbage collection messages                           */
    if ( 1 <= libGAP_SyMsgsFlagBags && full ) {
        if ( phase == 0 ) { libGAP_SyFputs( "#G  FULL ", 3 );                     }
        if ( phase == 1 ) { libGAP_SyFputs( str, 3 );  libGAP_SyFputs( "/",         3 ); }
        if ( phase == 2 ) { libGAP_SyFputs( str, 3 );  libGAP_SyFputs( shifted ? "mb live  " : "kb live  ", 3 ); }
        if ( phase == 3 ) { libGAP_SyFputs( str, 3 );  libGAP_SyFputs( "/",         3 ); }
        if ( phase == 4 ) { libGAP_SyFputs( str, 3 );  libGAP_SyFputs( shifted ? "mb dead  " : "kb dead  ", 3 ); }
        if ( phase == 5 ) { libGAP_SyFputs( str, 3 );  libGAP_SyFputs( "/",         3 ); }
        if ( phase == 6 ) { libGAP_SyFputs( str, 3 );  libGAP_SyFputs( shifted ? "mb free\n" : "kb free\n", 3 ); }
    }

    /* ordinary partial garbage collection messages                        */
    if ( 2 <= libGAP_SyMsgsFlagBags && ! full ) {
        if ( phase == 0 ) { libGAP_SyFputs( "#G  PART ", 3 );                     }
        if ( phase == 1 ) { libGAP_SyFputs( str, 3 );  libGAP_SyFputs( "/",         3 ); }
        if ( phase == 2 ) { libGAP_SyFputs( str, 3 );  libGAP_SyFputs( shifted ? "mb+live  ":"kb+live  ", 3 ); }
        if ( phase == 3 ) { libGAP_SyFputs( str, 3 );  libGAP_SyFputs( "/",         3 ); }
        if ( phase == 4 ) { libGAP_SyFputs( str, 3 );  libGAP_SyFputs( shifted ? "mb+dead  ":"kb+dead  ", 3 ); }
        if ( phase == 5 ) { libGAP_SyFputs( str, 3 );  libGAP_SyFputs( "/",         3 ); }
        if ( phase == 6 ) { libGAP_SyFputs( str, 3 );  libGAP_SyFputs( shifted ? "mb free\n":"kb free\n", 3 ); }
    }
    /* package (window) mode full garbage collection messages              */
    if ( phase != 0 ) {
      shifted =   3 <= phase && nr >= (1 << 21);
      if (shifted)
        nr *= 1024;
      cmd[0] = '@';
      cmd[1] = ( full ? '0' : ' ' ) + phase;
      cmd[2] = '\0';
      i = 0;
      for ( ; 0 < nr; nr /=10 )
        str[i++] = '0' + (nr % 10);
      str[i++] = '+';
      str[i++] = '\0';
      if (shifted)
        str[i++] = 'k';
      libGAP_syWinPut( 1, cmd, str );
    }
}



/****************************************************************************
**
*F  SyAllocBags( <size>, <need> ) . . . allocate memory block of <size> kilobytes
**
**  'SyAllocBags' is called from Gasman to get new storage from the operating
**  system.  <size> is the needed amount in kilobytes (it is always a multiple of
**  512 KByte),  and <need> tells 'SyAllocBags' whether  Gasman  really needs
**  the storage or only wants it to have a reasonable amount of free storage.
**
**  Currently  Gasman  expects this function to return  immediately  adjacent
**  areas on subsequent calls.  So 'sbrk' will  work  on  most  systems,  but
**  'malloc' will not.
**
**  If <need> is 0, 'SyAllocBags' must return 0 if it cannot or does not want
**  to extend the workspace,  and a pointer to the allocated area to indicate
**  success.   If <need> is 1  and 'SyAllocBags' cannot extend the workspace,
**  'SyAllocBags' must abort,  because GAP assumes that  'NewBag'  will never
**  fail.
**
**  <size> may also be negative in which case 'SyAllocBags' should return the
**  storage to the operating system.  In this case  <need>  will always be 0.
**  'SyAllocBags' can either accept this reduction and  return 1  and  return
**  the storage to the operating system or refuse the reduction and return 0.
**
**  If the operating system does not support dynamic memory managment, simply
**  give 'SyAllocBags' a static buffer, from where it returns the blocks.
*/

static libGAP_UInt libGAP_SyAllocPool;

/****************************************************************************
**
*f  SyAllocBags( <size>, <need> )
**
**  For UNIX, 'SyAllocBags' calls 'sbrk', which will work on most systems.
**
**  Note that   it may  happen that  another   function   has  called  'sbrk'
**  between  two calls to  'SyAllocBags',  so that the  next  allocation will
**  not be immediately adjacent to the last one.   In this case 'SyAllocBags'
**  returns the area to the operating system,  and either returns 0 if <need>
**  was 0 or aborts GAP if <need> was 1.  'SyAllocBags' will refuse to extend
**  the workspace beyond 'SyStorMax' or to reduce it below 'SyStorMin'.
*/

static libGAP_UInt libGAP_pagesize = 4096;   /* Will be initialised if SyAllocPool > 0 */

static inline libGAP_UInt libGAP_SyRoundUpToPagesize(libGAP_UInt x)
{
    libGAP_UInt r;
    r = x % libGAP_pagesize;
    return r == 0 ? x : x - r + libGAP_pagesize;
}

void *     libGAP_POOL = NULL;
libGAP_UInt * * * libGAP_syWorkspace = NULL;
libGAP_UInt       libGAP_syWorksize = 0;


#if HAVE_MADVISE
#ifndef libGAP_MAP_ANONYMOUS
#define libGAP_MAP_ANONYMOUS MAP_ANON
#endif

static void *libGAP_SyMMapStart = NULL;   /* Start of mmap'ed region for POOL */
static void *libGAP_SyMMapEnd;            /* End of mmap'ed region for POOL */
static void *libGAP_SyMMapAdvised;        /* We have already advised about non-usage
                                      up to here. */

void libGAP_SyMAdviseFree() {
    size_t size;
    void *from;
    if (!libGAP_SyMMapStart) 
        return;
    from = (char *) libGAP_syWorkspace + libGAP_syWorksize * 1024;
    from = (void *)libGAP_SyRoundUpToPagesize((libGAP_UInt) from);
    if (from > libGAP_SyMMapAdvised) {
        libGAP_SyMMapAdvised = from;
        return;
    }
    if (from < libGAP_SyMMapStart || from >= libGAP_SyMMapEnd || from >= libGAP_SyMMapAdvised)
        return;
    size = (char *)libGAP_SyMMapAdvised - (char *)from;
#if defined(MADV_FREE)
    madvise(from, size, MADV_FREE);
#elif defined(MADV_DONTNEED)
    madvise(from, size, MADV_DONTNEED);
#endif
    libGAP_SyMMapAdvised = from;
    /* On Darwin, MADV_FREE and MADV_DONTNEED will not actually update
     * a process's resident memory until those pages are explicitly
     * unmapped or needed elsewhere.
     *
     * The following code accomplishes this, but is not portable and
     * potentially not safe, since the POSIX standard does not make
     * any sufficiently strong promises with regard to the use of
     * MAP_FIXED.
     *
     * We probably don't want to do this and just live with pages
     * remaining with a process until reused even if that appears to
     * inflate the resident set size.
     *
     * Maybe we do want to do this until it breaks to avoid questions
     * by users...
     */
#ifndef NO_DIRTY_OSX_MMAP_TRICK
#if SYS_IS_DARWIN
    if (mmap(from, size, PROT_NONE,
            MAP_PRIVATE|libGAP_MAP_ANONYMOUS|MAP_FIXED, -1, 0) != from) {
        fputs("gap: OS X trick to free pages did not work, bye!\n", stderr);
        libGAP_SyExit( 2 );
    }
    if (mmap(from, size, PROT_READ|PROT_WRITE,
            MAP_PRIVATE|libGAP_MAP_ANONYMOUS|MAP_FIXED, -1, 0) != from) {
        fputs("gap: OS X trick to free pages did not work, bye!\n", stderr);
        libGAP_SyExit( 2 );
    }
#endif
#endif
}

void *libGAP_SyAnonMMap(size_t size) {
    void *result;
    size = libGAP_SyRoundUpToPagesize(size);
#ifdef libGAP_SYS_IS_64_BIT
    /* The following is at 16 Terabyte: */
    result = mmap((void *) (16L*1024*1024*1024*1024), size, 
                  PROT_READ|PROT_WRITE, MAP_PRIVATE|libGAP_MAP_ANONYMOUS, -1, 0);
    if (result == MAP_FAILED) {
        result = mmap(NULL, size, PROT_READ|PROT_WRITE,
            MAP_PRIVATE|libGAP_MAP_ANONYMOUS, -1, 0);
    }
#else
    result = mmap(NULL, size, PROT_READ|PROT_WRITE,
        MAP_PRIVATE|libGAP_MAP_ANONYMOUS, -1, 0);
#endif
    if (result == MAP_FAILED)
        result = NULL;
    libGAP_SyMMapStart = result;
    libGAP_SyMMapEnd = (char *)result + size;
    libGAP_SyMMapAdvised = (char *)result + size;
    return result;
}

int libGAP_SyTryToIncreasePool(void)
/* This tries to increase the pool size by a factor of 3/2, if this
 * worked, then 0 is returned, otherwise -1. */
{
    void *result;
    size_t size;
    size_t newchunk;

    size = (libGAP_Int) libGAP_SyMMapEnd - (libGAP_Int) libGAP_SyMMapStart;
    newchunk = libGAP_SyRoundUpToPagesize(size/2);
    result = mmap(libGAP_SyMMapEnd, newchunk, PROT_READ|PROT_WRITE,
                  MAP_PRIVATE|libGAP_MAP_ANONYMOUS, -1, 0);
    if (result == MAP_FAILED) return -1;
    if (result != libGAP_SyMMapEnd) {
        munmap(result,newchunk);
        return -1;
    }
    /* We actually got an extension! */
    libGAP_SyMMapEnd = (void *)((char *)libGAP_SyMMapEnd + newchunk);
    libGAP_SyAllocPool += newchunk;
    return 0;
}

#else

void libGAP_SyMAdviseFree(void) {
    /* do nothing */
}

int libGAP_SyTryToIncreasePool(void)
{
    return -1;   /* Refuse */
}

#endif

int libGAP_halvingsdone = 0;

void libGAP_SyInitialAllocPool( void )
{
#if HAVE_SYSCONF
#ifdef _SC_PAGESIZE
   libGAP_pagesize = sysconf(_SC_PAGESIZE);
#endif
#endif
   /* Otherwise we take the default of 4k as pagesize. */

   do {
       /* Always round up to pagesize: */
       libGAP_SyAllocPool = libGAP_SyRoundUpToPagesize(libGAP_SyAllocPool);
#if HAVE_MADVISE
       libGAP_POOL = libGAP_SyAnonMMap(libGAP_SyAllocPool+libGAP_pagesize);   /* For alignment */
#else
       libGAP_POOL = calloc(libGAP_SyAllocPool+libGAP_pagesize,1);   /* For alignment */
#endif
       if (libGAP_POOL != NULL) {
           /* fprintf(stderr,"Pool size is %lx.\n",SyAllocPool); */
           break;
       }
       libGAP_SyAllocPool = libGAP_SyAllocPool / 2;
       libGAP_halvingsdone++;
       if (libGAP_SyDebugLoading) fputs("gap: halving pool size.\n", stderr);
       if (libGAP_SyAllocPool < 16*1024*1024) {
         fputs("gap: cannot allocate initial memory, bye.\n", stderr);
         libGAP_SyExit( 2 );
       }
   } while (1);   /* Is left by break */

   /* ensure alignment of start address */
   libGAP_syWorkspace = (libGAP_UInt***)(libGAP_SyRoundUpToPagesize((libGAP_UInt) libGAP_POOL));
   /* Now both syWorkspace and SyAllocPool are aligned to pagesize */
}

libGAP_UInt ***libGAP_SyAllocBagsFromPool(libGAP_Int size, libGAP_UInt need)
{
  /* get the storage, but only if we stay within the bounds              */
  /* if ( (0 < size && syWorksize + size <= SyStorMax) */
  /* first check if we would get above SyStorKill, if yes exit! */
  if ( need < 2 && libGAP_SyStorKill != 0 && 0 < size 
                && libGAP_SyStorKill < libGAP_syWorksize + size ) {
      fputs("gap: will not extend workspace above -K limit, bye!\n",stderr);
      libGAP_SyExit( 2 );
  }
  if (size > 0) {
    while ((libGAP_syWorksize+size)*1024 > libGAP_SyAllocPool) {
        if (libGAP_SyTryToIncreasePool()) return (libGAP_UInt***)-1;
    }
    return (libGAP_UInt***)((char*)libGAP_syWorkspace + libGAP_syWorksize*1024);
  }
  else if  (size < 0 && (need >= 2 || libGAP_SyStorMin <= libGAP_syWorksize + size))
    return (libGAP_UInt***)((char*)libGAP_syWorkspace + libGAP_syWorksize*1024);
  else
    return (libGAP_UInt***)-1;
}

#if HAVE_SBRK && ! HAVE_VM_ALLOCATE /* prefer `vm_allocate' over `sbrk' */

libGAP_UInt * * * libGAP_SyAllocBags (
    libGAP_Int                 size,
    libGAP_UInt                need )
{
    libGAP_UInt * * *          ret;
    libGAP_UInt adjust = 0;

    if (libGAP_SyAllocPool > 0) {
      if (libGAP_POOL == NULL) libGAP_SyInitialAllocPool();
      /* Note that this does abort GAP if it does not succeed! */
      
      ret = libGAP_SyAllocBagsFromPool(size,need);
    }
    else {



        /* force alignment on first call                                       */
        if ( libGAP_syWorkspace == (libGAP_UInt***)0 ) {
#ifdef libGAP_SYS_IS_64_BIT
            libGAP_syWorkspace = (libGAP_UInt***)sbrk( 8 - (libGAP_UInt)sbrk(0) % 8 );
#else
            libGAP_syWorkspace = (libGAP_UInt***)sbrk( 4 - (libGAP_UInt)sbrk(0) % 4 );
#endif
            libGAP_syWorkspace = (libGAP_UInt***)sbrk( 0 );
        }

        /* get the storage, but only if we stay within the bounds              */
        /* if ( (0 < size && syWorksize + size <= SyStorMax) */
        /* first check if we would get above SyStorKill, if yes exit! */
        if ( need < 2 && libGAP_SyStorKill != 0 && 0 < size && 
             libGAP_SyStorKill < libGAP_syWorksize + size ) {
            fputs("gap: will not extend workspace above -K limit, bye!\n",
                  stderr);
            libGAP_SyExit( 2 );
        }
        if (0 < size )
          {
#ifndef libGAP_SYS_IS_64_BIT
            while (size > 1024*1024)
              {
                ret = (libGAP_UInt ***)sbrk(1024*1024*1024);
                if (ret != (libGAP_UInt ***)-1  && 
                    ret != (libGAP_UInt***)((char*)libGAP_syWorkspace + libGAP_syWorksize*1024))
                  {
                    sbrk(-1024*1024*1024);
                    ret = (libGAP_UInt ***)-1;
                  }
                if (ret == (libGAP_UInt ***)-1)
                  break;
                memset((void *)((char *)libGAP_syWorkspace + libGAP_syWorksize*1024), 0, 
                       1024*1024*1024);
                size -= 1024*1024;
                libGAP_syWorksize += 1024*1024;
                adjust++;
              }
#endif
            ret = (libGAP_UInt ***)sbrk(size*1024);
            if (ret != (libGAP_UInt ***)-1  && 
                ret != (libGAP_UInt***)((char*)libGAP_syWorkspace + libGAP_syWorksize*1024))
              {
                sbrk(-size*1024);
                ret = (libGAP_UInt ***)-1;
              }
            if (ret != (libGAP_UInt ***)-1)
              memset((void *)((char *)libGAP_syWorkspace + libGAP_syWorksize*1024), 0, 
                     1024*size);
            
          }
        else if  (size < 0 && (need >= 2 || libGAP_SyStorMin <= libGAP_syWorksize + size))  {
#ifndef libGAP_SYS_IS_64_BIT
          while (size < -1024*1024)
            {
              ret = (libGAP_UInt ***)sbrk(-1024*1024*1024);
              if (ret == (libGAP_UInt ***)-1)
                break;
              size += 1024*1024;
              libGAP_syWorksize -= 1024*1024;
            }
#endif
            ret = (libGAP_UInt ***)sbrk(size*1024);
        }
        else {
          ret = (libGAP_UInt***)-1;
        }
    }


    /* update the size info                                                */
    if ( ret != (libGAP_UInt***)-1 ) {
        libGAP_syWorksize += size;
        /* set the overrun flag if we became larger than SyStorMax */
        if ( libGAP_SyStorMax != 0 && libGAP_syWorksize  > libGAP_SyStorMax)  {
          libGAP_SyStorOverrun = -1;
          libGAP_SyStorMax=libGAP_syWorksize*2; /* new maximum */
          libGAP_InterruptExecStat(); /* interrupt at the next possible point */
        }
    }

    /* test if the allocation failed                                       */
    if ( ret == (libGAP_UInt***)-1 && need ) {
        fputs("gap: cannot extend the workspace any more!\n",stderr);
        libGAP_SyExit( 1 );
    }
    /* if we de-allocated the whole workspace then remember this */
    if (libGAP_syWorksize == 0)
      libGAP_syWorkspace = (libGAP_UInt ***)0;

    /* otherwise return the result (which could be 0 to indicate failure)  */
    if ( ret == (libGAP_UInt***)-1 )
        return 0;
    else
      {
        return (libGAP_UInt***)(((libGAP_Char *)ret) - 1024*1024*1024*adjust);
      }

}

#endif


/****************************************************************************
**
*f  SyAllocBags( <size>, <need> )
**
**  Under MACH virtual memory managment functions are used instead of 'sbrk'.
*/
#if HAVE_VM_ALLOCATE

#include <mach/mach.h>

#if (defined(SYS_IS_DARWIN) && SYS_IS_DARWIN) || defined(__gnu_hurd__)
#define libGAP_task_self mach_task_self
#endif

vm_address_t libGAP_syBase;
 
libGAP_UInt * * * libGAP_SyAllocBags (
    libGAP_Int                 size,
    libGAP_UInt                need )
{
    libGAP_UInt * * *          ret = (libGAP_UInt***)-1;
    vm_address_t        adr;    

    if (libGAP_SyAllocPool > 0) {
      if (libGAP_POOL == NULL) libGAP_SyInitialAllocPool();
      /* Note that this does abort GAP if it does not succeed! */
 
      ret = libGAP_SyAllocBagsFromPool(size,need);
      if (ret != (libGAP_UInt ***)-1)
          libGAP_syWorksize += size;

    }
    else {
        if ( libGAP_SyStorKill != 0 && 0 < size && libGAP_SyStorKill < 1024*(libGAP_syWorksize + size) ) {
            if (need) {
                fputs("gap: will not extend workspace above -K limit, bye!\n",stderr);
                libGAP_SyExit( 2 );
            }  
        }
        /* check that <size> is divisible by <vm_page_size>                    */
        else if ( size*1024 % vm_page_size != 0 ) {
            fputs( "gap: memory block size is not a multiple of vm_page_size",
                   stderr );
            libGAP_SyExit(1);
        }

        /* check that we don't try to shrink uninialized memory                */
        else if ( size <= 0 && libGAP_syBase == 0 ) {
            fputs( "gap: trying to shrink uninialized vm memory\n", stderr );
            libGAP_SyExit(1);
        }

        /* allocate memory anywhere on first call                              */
        else if ( 0 < size && libGAP_syBase == 0 ) {
            if ( vm_allocate(libGAP_task_self(),&libGAP_syBase,size*1024,TRUE) == KERN_SUCCESS ) {
                libGAP_syWorksize = size;
                ret = (libGAP_UInt***) libGAP_syBase;
            }
        }

        /* don't shrink memory but mark it as deactivated                      */
        else if ( size < 0 && libGAP_syWorksize + size > libGAP_SyStorMin) {
            adr = (vm_address_t)( (char*) libGAP_syBase + (libGAP_syWorksize+size)*1024 );
            if ( vm_deallocate(libGAP_task_self(),adr,-size*1024) == KERN_SUCCESS ) {
                ret = (libGAP_UInt***)( (char*) libGAP_syBase + libGAP_syWorksize*1024 );
                libGAP_syWorksize += size;
            }
        }

        /* get more memory from system                                         */
        else {
            adr = (vm_address_t)( (char*) libGAP_syBase + libGAP_syWorksize*1024 );
            if ( vm_allocate(libGAP_task_self(),&adr,size*1024,FALSE) == KERN_SUCCESS ) {
                ret = (libGAP_UInt***) ( (char*) libGAP_syBase + libGAP_syWorksize*1024 );
                libGAP_syWorksize += size;
            }
        }

        /* test if the allocation failed                                       */
        if ( ret == (libGAP_UInt***)-1 && need ) {
            fputs("gap: cannot extend the workspace any more!!\n",stderr);
            libGAP_SyExit(1);
        }
    }

    /* otherwise return the result (which could be 0 to indicate failure)  */
    if ( ret == (libGAP_UInt***)-1 ){
        if (need) { 
            fputs("gap: cannot extend the workspace any more!!!\n",stderr);
            libGAP_SyExit( 1 );
        }
        return (libGAP_UInt***) 0;
    } 
    else {
        if (libGAP_syWorksize  > libGAP_SyStorMax)  {
            libGAP_SyStorOverrun = -1;
            libGAP_SyStorMax=libGAP_syWorksize*2; /* new maximum */
            libGAP_InterruptExecStat(); /* interrupt at the next possible point */
       }
     }
    return ret;
}

#endif


/****************************************************************************
**
*F  SyAbortBags( <msg> )  . . . . . . . . . abort GAP in case of an emergency
**
**  'SyAbortBags' is the function called by Gasman in case of an emergency.
*/
void libGAP_SyAbortBags (
    const libGAP_Char *        msg )
{
    libGAP_SyFputs( msg, 3 );
    libGAP_SyExit( 2 );
}

/****************************************************************************
**
*F  SySleep( <secs> ) . . . . . . . . . . . . . .sleep GAP for <secs> seconds
**
**  NB Various OS events (like signals) might wake us up
**
*/
void libGAP_SySleep ( libGAP_UInt secs )
{
  sleep( (unsigned int) secs );
}

/****************************************************************************
**

*F * * * * * * * * * * * * * initialize package * * * * * * * * * * * * * * *
*/


/****************************************************************************
**

*F  SyExit( <ret> ) . . . . . . . . . . . . . exit GAP with return code <ret>
**
**  'SyExit' is the offical  way  to  exit GAP, bus errors are the inoffical.
**  The function 'SyExit' must perform all the neccessary cleanup operations.
**  If ret is 0 'SyExit' should signal to a calling proccess that all is  ok.
**  If ret is 1 'SyExit' should signal a  failure  to  the  calling proccess.
*/
void libGAP_SyExit (
    libGAP_UInt                ret )
{
#if SYS_IS_CYGWIN32
  if (ret!=0) {
    libGAP_Int c;
    fputs("gap: Press <Enter> to end program\n",stderr);
    do {
      c=libGAP_SyGetch(1);   /* wait for the user to type <return> */
    } while (c!='\n' && c!=' ');
  }

#endif

    exit( (int)ret );
}

/****************************************************************************
**
*F  SySetGapRootPath( <string> )  . . . . . . . . .  set the root directories
**
**  'SySetGapRootPath' takes a string and modifies a list of root directories
**  in 'SyGapRootPaths'.
**
**  A  leading semicolon in  <string> means  that the list  of directories in
**  <string> is  appended  to the  existing list  of  root paths.  A trailing
**  semicolon means they are prepended.   If there is  no leading or trailing
**  semicolon, then the root paths are overwritten.
*/


/****************************************************************************
**
*f  SySetGapRootPath( <string> )
*/
#if HAVE_SLASH_SEPARATOR 

void libGAP_SySetGapRootPath( const libGAP_Char * string )
{
    const libGAP_Char *          p;
    libGAP_Char *          q;
    libGAP_Int             i;
    libGAP_Int             n;

    /* set string to a default value if unset                              */
    if ( string == 0 || *string == 0 ) {
        string = "./";
    }
 
    /* 
    ** check if we append, prepend or overwrite. 
    */ 
    if( string[0] == ';' ) {
        /* Count the number of root directories already present.           */
         n = 0; while( libGAP_SyGapRootPaths[n][0] != '\0' ) n++;

         /* Skip leading semicolon.                                        */
         string++;

    }
    else if( string[ strlen(string) - 1 ] == ';' ) {
        /* Count the number of directories in 'string'.                    */
        n = 0; p = string; while( *p ) if( *p++ == ';' ) n++;

        /* Find last root path.                                            */
        for( i = 0; i < libGAP_MAX_GAP_DIRS; i++ ) 
            if( libGAP_SyGapRootPaths[i][0] == '\0' ) break;
        i--;

        /* Move existing root paths to the back                            */
        if( i + n >= libGAP_MAX_GAP_DIRS ) return;
        while( i >= 0 ) {
            memcpy( libGAP_SyGapRootPaths[i+n], libGAP_SyGapRootPaths[i], sizeof(libGAP_SyGapRootPaths[i+n]) );
            i--;
        }

        n = 0;

    }
    else {
        /* Make sure to wipe out all possibly existing root paths          */
        for( i = 0; i < libGAP_MAX_GAP_DIRS; i++ ) libGAP_SyGapRootPaths[i][0] = '\0';
        n = 0;
    }

    /* unpack the argument                                                 */
    p = string;
    while ( *p ) {
        if( n >= libGAP_MAX_GAP_DIRS ) return;

        q = libGAP_SyGapRootPaths[n];
        while ( *p && *p != ';' ) {
            *q = *p++;

#if SYS_IS_CYGWIN32
            /* fix up for DOS */
            if (*q == '\\')
              *q = '/';
#endif
            
            q++;
        }
        if ( q == libGAP_SyGapRootPaths[n] ) {
            libGAP_strxcpy( libGAP_SyGapRootPaths[n], "./", sizeof(libGAP_SyGapRootPaths[n]) );
        }
        else if ( q[-1] != '/' ) {
            *q++ = '/';
            *q   = '\0';
        }
        else {
            *q   = '\0';
        }
        if ( *p ) {
            p++;  n++;
        }
    }
    return; 
}

#endif


/****************************************************************************
**
*F  InitSystem( <argc>, <argv> )  . . . . . . . . . initialize system package
**
**  'InitSystem' is called very early during the initialization from  'main'.
**  It is passed the command line array  <argc>, <argv>  to look for options.
**
**  For UNIX it initializes the default files 'stdin', 'stdout' and 'stderr',
**  installs the handler 'syAnsIntr' to answer the user interrupts '<ctr>-C',
**  scans the command line for options, tries to  find  'LIBNAME/init.g'  and
**  '$HOME/.gaprc' and copies the remaining arguments into 'SyInitfiles'.
*/

typedef struct { libGAP_Char symbol; libGAP_UInt value; } libGAP_sizeMultiplier;

libGAP_sizeMultiplier libGAP_memoryUnits[]= {
  {'k', 1024},
  {'K', 1024},
  {'m', 1024*1024},
  {'M', 1024*1024},
  {'g', 1024*1024*1024},
  {'G', 1024*1024*1024},
#ifdef libGAP_SYS_IS_64_BIT
  {'t', 1024UL*1024*1024*1024},
  {'T', 1024UL*1024*1024*1024},
  {'p', 1024UL*1024*1024*1024*1024}, /* you never know */
  {'P', 1024UL*1024*1024*1024*1024},
#endif
};

static libGAP_UInt libGAP_ParseMemory( libGAP_Char * s)
{
  libGAP_UInt size  = atol(s);
  libGAP_Char symbol =  s[strlen(s)-1];
  libGAP_UInt i;
  libGAP_UInt maxmem;
#ifdef libGAP_SYS_IS_64_BIT
  maxmem = 15000000000000000000UL;
#else
  maxmem = 4000000000UL;
#endif
  
  for (i = 0; i < sizeof(libGAP_memoryUnits)/sizeof(libGAP_memoryUnits[0]); i++) {
    if (symbol == libGAP_memoryUnits[i].symbol) {
      libGAP_UInt value = libGAP_memoryUnits[i].value;
      if (size > maxmem/value)
        size = maxmem;
      else
        size *= value;
      return size;
    }      
  }
  if (!libGAP_IsDigit(symbol))
    libGAP_FPUTS_TO_STDERR("Unrecognised memory unit ignored");
  return size;
}


struct libGAP_optInfo {
  libGAP_Char key;
  libGAP_Int (*handler)(libGAP_Char **, void *);
  void *otherArg;
  libGAP_UInt minargs;
};


static libGAP_Int libGAP_toggle( libGAP_Char ** argv, void *Variable )
{
  libGAP_UInt * variable = (libGAP_UInt *) Variable;
  *variable = !*variable;
  return 0;
}

static libGAP_Int libGAP_storeString( libGAP_Char **argv, void *Where )
{
  libGAP_Char **where = (libGAP_Char **)Where;
  *where = argv[0];
  return 1;
}

static libGAP_Int libGAP_storeMemory( libGAP_Char **argv, void *Where )
{
  libGAP_UInt *where = (libGAP_UInt *)Where;
  *where = libGAP_ParseMemory(argv[0]);
  return 1;
}

static libGAP_Int libGAP_storeMemory2( libGAP_Char **argv, void *Where )
{
  libGAP_UInt *where = (libGAP_UInt *)Where;
  *where = libGAP_ParseMemory(argv[0])/1024;
  return 1;
}

static libGAP_Int libGAP_processCompilerArgs( libGAP_Char **argv, void * dummy)
{
  libGAP_SyCompilePlease = 1;
  libGAP_strxcat( libGAP_SyCompileOutput, argv[0], sizeof(libGAP_SyCompileOutput) );
  libGAP_strxcat( libGAP_SyCompileInput, argv[1], sizeof(libGAP_SyCompileInput) );
  libGAP_strxcat( libGAP_SyCompileName, argv[2], sizeof(libGAP_SyCompileName) );
  libGAP_SyCompileMagic1 = argv[3];
  return 4;
}

static libGAP_Int libGAP_unsetString( libGAP_Char **argv, void *Where)
{
  *(libGAP_Char **)Where = (libGAP_Char *)0;
  return 0;
}

static libGAP_Int libGAP_forceLineEditing( libGAP_Char **argv,void *Level)
{
  libGAP_UInt level = (libGAP_UInt)Level;
  libGAP_SyLineEdit = level;
  return 0;
}

static libGAP_Int libGAP_setGapRootPath( libGAP_Char **argv, void *Dummy)
{
  libGAP_SySetGapRootPath( argv[0] );
  return 1;
}


static libGAP_Int libGAP_preAllocAmount;

/* These are just the options that need kernel processing. Additional options will be 
   recognised and handled in the library */

struct libGAP_optInfo libGAP_options[] = {
  { 'B',  libGAP_storeString, &libGAP_SyArchitecture, 1}, /* default architecture needs to be passed from kernel 
                                               to library. Might be needed for autoload of compiled files */
  { 'C',  libGAP_processCompilerArgs, 0, 4}, /* must handle in kernel */
  { 'D',  libGAP_toggle, &libGAP_SyDebugLoading, 0}, /* must handle in kernel */
  { 'K',  libGAP_storeMemory2, &libGAP_SyStorKill, 1}, /* could handle from library with new interface */
  { 'L',  libGAP_storeString, &libGAP_SyRestoring, 1}, /* must be handled in kernel  */
  { 'M',  libGAP_toggle, &libGAP_SyUseModule, 0}, /* must be handled in kernel */
  { 'X',  libGAP_toggle, &libGAP_SyCheckCRCCompiledModule, 0}, /* must be handled in kernel */
  { 'R',  libGAP_unsetString, &libGAP_SyRestoring, 0}, /* kernel */
  { 'U',  libGAP_storeString, libGAP_SyCompileOptions, 1}, /* kernel */
  { 'a',  libGAP_storeMemory, &libGAP_preAllocAmount, 1 }, /* kernel -- is this still useful */
  { 'c',  libGAP_storeMemory, &libGAP_SyCacheSize, 1 }, /* kernel, unless we provided a hook to set it from library, 
                                           never seems to be useful */
  { 'e',  libGAP_toggle, &libGAP_SyCTRD, 0 }, /* kernel */
  { 'f',  libGAP_forceLineEditing, (void *)2, 0 }, /* probably library now */
  { 'E',  libGAP_toggle, &libGAP_SyUseReadline, 0 }, /* kernel */
  { 'i',  libGAP_storeString, libGAP_SySystemInitFile, 1}, /* kernel */
  { 'l',  libGAP_setGapRootPath, 0, 1}, /* kernel */
  { 'm',  libGAP_storeMemory2, &libGAP_SyStorMin, 1 }, /* kernel */
  { 'r',  libGAP_toggle, &libGAP_IgnoreGapRC, 0 }, /* kernel */
  { 's',  libGAP_storeMemory, &libGAP_SyAllocPool, 1 }, /* kernel */
  { 'n',  libGAP_forceLineEditing, 0, 0}, /* prob library */
  { 'o',  libGAP_storeMemory2, &libGAP_SyStorMax, 1 }, /* library with new interface */
  { 'p',  libGAP_toggle, &libGAP_SyWindow, 0 }, /* ?? */
  { 'q',  libGAP_toggle, &libGAP_SyQuiet, 0 }, /* ?? */
  { '\0',0,0}};


libGAP_Char ** libGAP_SyOriginalArgv;
libGAP_UInt libGAP_SyOriginalArgc;

 

void libGAP_InitSystem (
    libGAP_Int                 argc,
    libGAP_Char *              argv [] )
{
    libGAP_Char *              *ptrlist;
    libGAP_UInt                i;             /* loop variable                   */
    libGAP_Int res;                       /* return from option processing function */

    /* Initialize global and static variables. Do it here rather than
       with initializers to allow for restart */
    /* SyBanner = 1; */
    libGAP_SyCTRD = 1;             
    libGAP_SyCacheSize = 0;
    libGAP_SyCheckCRCCompiledModule = 0;
    libGAP_SyCompilePlease = 0;
    libGAP_SyDebugLoading = 0;
    libGAP_SyHasUserHome = 0;
    libGAP_SyLineEdit = 1;
    libGAP_SyUseReadline = 1;
    /* SyAutoloadPackages = 1; */
    /*  SyBreakSuppress = 0; */
    libGAP_SyMsgsFlagBags = 0;
    libGAP_SyNrCols = 0;
    libGAP_SyNrColsLocked = 0;
    libGAP_SyNrRows = 0;
    libGAP_SyNrRowsLocked = 0;
    libGAP_SyQuiet = 0;
    libGAP_SyInitializing = 0;
#ifdef libGAP_SYS_IS_64_BIT
    libGAP_SyStorMax = 2048*1024L;          /* This is in kB! */
    libGAP_SyAllocPool = 4096L*1024*1024;   /* Note this is in bytes! */
#else
    libGAP_SyStorMax = 1024*1024L;          /* This is in kB! */
#if SYS_IS_CYGWIN32
    libGAP_SyAllocPool = 0;                 /* works better on cygwin */
#else
    libGAP_SyAllocPool = 1536L*1024*1024;   /* Note this is in bytes! */
#endif
#endif
    libGAP_SyStorOverrun = 0;
    libGAP_SyStorKill = 0;
    libGAP_SyStorMin = libGAP_SY_STOR_MIN;         /* see system.h */
    libGAP_SyUseModule = 1;
    libGAP_SyWindow = 0;

    for (i = 0; i < 2; i++) {
      libGAP_UInt j;
      for (j = 0; j < 7; j++) {
        libGAP_SyGasmanNumbers[i][j] = 0;
      }
    }

#if HAVE_VM_ALLOCATE
    libGAP_syBase = 0;
    libGAP_syWorksize = 0;
#elif HAVE_SBRK
    libGAP_syWorkspace = (libGAP_UInt ***)0;
#endif
    /*  nopts = 0;
    noptvals = 0;
    lenoptvalsbuff = 0;
    gaprc = 1; */

    libGAP_preAllocAmount = 4*1024*1024;
    
    /* open the standard files                                             */
#if HAVE_TTYNAME
    libGAP_syBuf[0].fp = fileno(stdin);
    libGAP_syBuf[0].bufno = -1;
    if ( isatty( fileno(stdin) ) ) {
        if ( isatty( fileno(stdout) )
          && ! strcmp( ttyname(fileno(stdin)), ttyname(fileno(stdout)) ) )
            libGAP_syBuf[0].echo = fileno(stdout);
        else
            libGAP_syBuf[0].echo = open( ttyname(fileno(stdin)), O_WRONLY );
        libGAP_syBuf[0].isTTY = 1;
    }
    else {
        libGAP_syBuf[0].echo = fileno(stdout);
        libGAP_syBuf[0].isTTY = 0;
    }
    libGAP_syBuf[1].echo = libGAP_syBuf[1].fp = fileno(stdout); 
    libGAP_syBuf[1].bufno = -1;
    if ( isatty( fileno(stderr) ) ) {
        if ( isatty( fileno(stdin) )
          && ! strcmp( ttyname(fileno(stdin)), ttyname(fileno(stderr)) ) )
            libGAP_syBuf[2].fp = fileno(stdin);
        else
            libGAP_syBuf[2].fp = open( ttyname(fileno(stderr)), O_RDONLY );
        libGAP_syBuf[2].echo = fileno(stderr);
        libGAP_syBuf[2].isTTY = 1;
    }
    else
      libGAP_syBuf[2].isTTY = 0;
    libGAP_syBuf[2].bufno = -1;
    libGAP_syBuf[3].fp = fileno(stderr);
    libGAP_syBuf[3].bufno = -1;
    setbuf(stdin, (char *)0);
    setbuf(stdout, (char *)0);
    setbuf(stderr, (char *)0);
#endif

    for (i = 4; i < sizeof(libGAP_syBuf)/sizeof(libGAP_syBuf[0]); i++)
      libGAP_syBuf[i].fp = -1;
    
    for (i = 0; i < sizeof(libGAP_syBuffers)/sizeof(libGAP_syBuffers[0]); i++)
          libGAP_syBuffers[i].inuse = 0;

#if HAVE_LIBREADLINE
    rl_initialize ();
#endif
    
    libGAP_SyInstallAnswerIntr();

    libGAP_SySystemInitFile[0] = '\0';
    libGAP_strxcpy( libGAP_SySystemInitFile, "lib/init.g", sizeof(libGAP_SySystemInitFile) );
#if SYS_IS_CYGWIN32
    libGAP_SySetGapRootPath( libGAP_SyWindowsPath );
#else

#ifdef SYS_DEFAULT_PATHS
    libGAP_SySetGapRootPath( SYS_DEFAULT_PATHS );
#else
    libGAP_SySetGapRootPath( "./" );
#endif

#endif

    /* save the original command line for export to GAP */
    libGAP_SyOriginalArgc = argc;
    libGAP_SyOriginalArgv = argv;

    /* scan the command line for options that we have to process in the kernel */
    /* we just scan the whole command line looking for the keys for the options we recognise */
    /* anything else will presumably be dealt with in the library */
    while ( argc > 1 )
      {
        if (argv[1][0] == '-' ) {

          if ( strlen(argv[1]) != 2 ) {
            libGAP_FPUTS_TO_STDERR("gap: sorry, options must not be grouped '");
            libGAP_FPUTS_TO_STDERR(argv[1]);  libGAP_FPUTS_TO_STDERR("'.\n");
            goto usage;
          }


          for (i = 0; libGAP_options[i].key != argv[1][1] && libGAP_options[i].key; i++)
            ;

        


          if (argc < 2 + libGAP_options[i].minargs)
            {
              libGAP_Char buf[2];
              libGAP_FPUTS_TO_STDERR("gap: option "); libGAP_FPUTS_TO_STDERR(argv[1]);
              libGAP_FPUTS_TO_STDERR(" requires at least ");
              buf[0] = libGAP_options[i].minargs + '0';
              buf[1] = '\0';
              libGAP_FPUTS_TO_STDERR(buf); libGAP_FPUTS_TO_STDERR(" arguments\n");
              goto usage;
            }
          if (libGAP_options[i].handler) {
            res = (*libGAP_options[i].handler)(argv+2, libGAP_options[i].otherArg);
            
            switch (res)
              {
              case -1: goto usage;
                /*            case -2: goto fullusage; */
              default: ;     /* fall through and continue */
              }
          }
          else
            res = libGAP_options[i].minargs;
          /*    recordOption(argv[1][1], res,  argv+2); */
          argv += 1 + res;
          argc -= 1 + res;
          
        }
        else {
          argv++;
          argc--;
        }
          
      }
    /* adjust SyUseReadline if no readline support available or for XGAP  */
#if !HAVE_LIBREADLINE
    libGAP_SyUseReadline = 0;
#endif
    if (libGAP_SyWindow) libGAP_SyUseReadline = 0;

    /* now that the user has had a chance to give -x and -y,
       we determine the size of the screen ourselves */
    libGAP_getwindowsize();

    /* fix max if it is lower than min                                     */
    if ( libGAP_SyStorMax != 0 && libGAP_SyStorMax < libGAP_SyStorMin ) {
        libGAP_SyStorMax = libGAP_SyStorMin;
    }

    /* fix pool size if it is given and lower than SyStorMax */
    if ( libGAP_SyAllocPool != 0 && libGAP_SyAllocPool < libGAP_SyStorMax * 1024) {
        libGAP_SyAllocPool = libGAP_SyStorMax * 1024;
    }

    /* when running in package mode set ctrl-d and line editing            */
    if ( libGAP_SyWindow ) {
      /*         SyLineEdit   = 1;
                 SyCTRD       = 1; */
        libGAP_syBuf[2].fp = fileno(stdin);  libGAP_syBuf[2].echo = fileno(stdout);
        libGAP_syBuf[3].fp = fileno(stdout);
        libGAP_syWinPut( 0, "@p", "1." );
    }
   

    if (libGAP_SyAllocPool == 0) {
      /* premalloc stuff                                                     */
      /* allocate in small chunks, and write something to them
       * (the GNU clib uses mmap for large chunks and give it back to the
       * system after free'ing; also it seems that memory is only really 
       * allocated (pagewise) when it is first used)                     */
      ptrlist = (libGAP_Char **)malloc((1+libGAP_preAllocAmount/1000)*sizeof(libGAP_Char*));
      for (i = 1; i*1000 < libGAP_preAllocAmount; i++) {
        ptrlist[i-1] = (libGAP_Char *)malloc( 1000 );
        if (ptrlist[i-1] != NULL) ptrlist[i-1][900] = 13;
      }
      for (i = 1; (i+1)*1000 < libGAP_preAllocAmount; i++) 
        if (ptrlist[i-1] != NULL) free(ptrlist[i-1]);
      free(ptrlist);
       
     /* ptr = (Char *)malloc( preAllocAmount );
      ptr1 = (Char *)malloc(4);
      if ( ptr != 0 )  free( ptr ); */
    }

    /* try to find 'LIBNAME/init.g' to read it upon initialization         */
    if ( libGAP_SyCompilePlease || libGAP_SyRestoring ) {
        libGAP_SySystemInitFile[0] = 0;
    }

    /* the compiler will *not* read in the .gaprc file                     
    if ( gaprc && ! ( SyCompilePlease || SyRestoring ) ) {
        sySetGapRCFile();
    }
    */

#if HAVE_DOTGAPRC
    /* the users home directory                                            */
    if ( getenv("HOME") != 0 ) {
        libGAP_strxcpy(libGAP_SyUserHome, getenv("HOME"), sizeof(libGAP_SyUserHome));
        libGAP_SyHasUserHome = 1;

        libGAP_strxcpy(libGAP_DotGapPath, getenv("HOME"), sizeof(libGAP_DotGapPath));
# if defined(SYS_IS_DARWIN) && SYS_IS_DARWIN
        libGAP_strxcat(libGAP_DotGapPath, "/Library/Preferences/GAP;", sizeof(libGAP_DotGapPath));
# elif defined(__CYGWIN__)
        libGAP_strxcat(libGAP_DotGapPath, "/_gap;", sizeof(libGAP_DotGapPath));
# else
        libGAP_strxcat(libGAP_DotGapPath, "/.gap;", sizeof(libGAP_DotGapPath));
# endif

        if (!libGAP_IgnoreGapRC) {
          libGAP_SySetGapRootPath(libGAP_DotGapPath);
        }
        libGAP_DotGapPath[strlen(libGAP_DotGapPath)-1] = '\0';
        
        /* and in this case we can also expand paths which start
           with a tilde ~ */
        for (i = 0; i < libGAP_MAX_GAP_DIRS && libGAP_SyGapRootPaths[i][0]; i++) {
          if (libGAP_SyGapRootPaths[i][0] == '~' && 
              strlen(libGAP_SyUserHome)+strlen(libGAP_SyGapRootPaths[i]) < 512) {
            memmove(libGAP_SyGapRootPaths[i]+strlen(libGAP_SyUserHome),
                    /* don't copy the ~ but the trailing '\0' */
                    libGAP_SyGapRootPaths[i]+1, strlen(libGAP_SyGapRootPaths[i]));
            memcpy(libGAP_SyGapRootPaths[i], libGAP_SyUserHome, strlen(libGAP_SyUserHome));
          }
        }
    }
#endif


    /* start the clock                                                     */
    libGAP_SyStartTime = libGAP_SyTime();



    /* now we start                                                        */
    return;

    /* print a usage message                                               */
usage:
 libGAP_FPUTS_TO_STDERR("usage: gap [OPTIONS] [FILES]\n");
 libGAP_FPUTS_TO_STDERR("       run the Groups, Algorithms and Programming system, Version ");
 libGAP_FPUTS_TO_STDERR(libGAP_SyKernelVersion);
 libGAP_FPUTS_TO_STDERR("\n");
 libGAP_FPUTS_TO_STDERR("       use '-h' option to get help.\n");
 libGAP_FPUTS_TO_STDERR("\n");
 libGAP_SyExit( 1 );
}


/****************************************************************************
**
*E  system.c  . . . . . . . . . . . . . . . . . . . . . . . . . . . ends here
*/

